/*-
 * Copyright (c) 2004 Gleb Smirnoff <glebius@cell.sick.ru>
 * Copyright (c) 2001-2003 Roman V. Palagin <romanp@unshadow.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * This product includes software developed by Gleb Smirnoff and
 * contributors.
 * 4. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	 $Id: netflow.c,v 1.11.2.10 2004/08/10 13:21:54 glebius Exp $
 */

static const char rcs_id[] =
    "@(#) $Id: netflow.c,v 1.11.2.10 2004/08/10 13:21:54 glebius Exp $";

#include <sys/param.h>
#include <sys/kernel.h>
#if __FreeBSD_version >= 500000
#include <sys/limits.h>
#else
#include <machine/limits.h>
#endif
#include <sys/mbuf.h>
#include <sys/systm.h>
#include <sys/socket.h>

#include <net/route.h>
#include <net/if.h>
#include <net/if_var.h>
#include <net/if_dl.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>

#define  _IP_VHL
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

#include <netgraph/ng_message.h>
#include <netgraph/netgraph.h>

#include "ng_netflow.h"
#include "netflow.h"
#include "kernel.h"

#define	NBUCKETS	(4096)	/* must be power of 2 */

/*
 * If you have better ideas about hashing function, please send
 * them to <glebius@cell.sick.ru>
 */

#define FULL_HASH(addr1,addr2,port1,port2)\
	(((addr1 >> 16) ^		\
	  (addr2 & 0x00FF) ^		\
	  ((port1 ^ port2) << 8) )&	\
	 (NBUCKETS - 1))

#define ADDR_HASH(addr1,addr2)\
	(((addr1 >> 16) ^		\
	  (addr2 & 0x00FF) )&		\
	 (NBUCKETS - 1))
	
MALLOC_DECLARE(M_NETFLOW);
MALLOC_DEFINE(M_NETFLOW, "NetFlow", "flow cache");

static int export_add(priv_p , struct flow_entry *);
static int export_send(priv_p );

/*
 * These are hash manipulating functions
 */

/* Generate hash */
static __inline u_int32_t
ip_hash(struct flow_rec *r)
{
	switch (r->r_ip_p) {
	case IPPROTO_TCP:
	case IPPROTO_UDP:
		return FULL_HASH(r->r_src.s_addr,r->r_dst.s_addr,r->r_sport,r->r_dport);
	default:
		return ADDR_HASH(r->r_src.s_addr,r->r_dst.s_addr);
	}
}

/* Lookup for record in defined slot */
static __inline struct flow_entry *
hash_lookup(struct flow_hash *h, int slot, struct flow_rec *r)
{
	struct flow_entry *fle;

	LIST_FOREACH(fle, &(h[slot].head), next) {
		if(bcmp(r, &fle->r, sizeof(struct flow_rec)) == 0) {
			return (fle);
		}
	}
	return (NULL);
}

/* Lookup for free flow_entry space in cache. Allocate more
 * memory if needed.
 */
static __inline struct flow_entry *
alloc_flow(priv_p priv)
{
	struct flow_chunk		*chunk;
	register struct flow_entry	*fle;
	register int	i;

	/* First lookup for cache chunk, which is not filled */
	LIST_FOREACH(chunk, &priv->cache, entries) {
		if (chunk->flows == CHUNK_SIZE ) {
			if (LIST_NEXT(chunk, entries) == NULL) {
				struct flow_chunk	*new;

				MALLOC(new, struct flow_chunk *, sizeof(struct flow_chunk),
				    M_NETFLOW, M_NOWAIT);
				if (new == NULL)
					return (NULL);	/* ENOMEM */

				bzero(new, sizeof(struct flow_chunk));
				LIST_INSERT_AFTER(chunk, new, entries);

				printf("ng_netflow: new chunk allocated\n");
			}
		} else
			break;
	}

	/* Then lookup for free space in chunk */
	for (i=0, fle = chunk->chunk; i < CHUNK_SIZE; i++, fle++)
		if (fle->packets == 0)	{
			chunk->flows ++;
			fle->my_chunk = chunk;
			return (fle);
		};



	return (NULL);	/* not reached */
}

/* Wipe flow_entry in cache */
static __inline void
free_flow(struct flow_entry *fle)
{
	fle->my_chunk->flows --;
	bzero(fle, sizeof(struct flow_entry));
}

/* Calculate number of bits in netmask */
#define	g21	0x55555555ul	/* = 0101_0101_0101_0101_0101_0101_0101_0101 */
#define	g22	0x33333333ul	/* = 0011_0011_0011_0011_0011_0011_0011_0011 */
#define	g23	0x0f0f0f0ful	/* = 0000_1111_0000_1111_0000_1111_0000_1111 */
static __inline u_char
bit_count(u_long v)
{
	v = (v & g21) + ((v >> 1) & g21);
	v = (v & g22) + ((v >> 2) & g22);
	v = (v + (v >> 4)) & g23;
	return (v + (v >> 8) + (v >> 16) + (v >> 24)) & 0x3f;
}

/* Insert record into defined slot.
 * Fill in all possible fields. */
static __inline int
hash_insert(priv_p priv, int slot, struct flow_rec *r, int plen, iface_p iface)
{
	struct flow_hash	*h = priv->hash;
	struct flow_entry	*fle;
	struct timeval tv;
	struct route ro;
	struct sockaddr_in *sin;

	fle = alloc_flow(priv);
	if (fle == NULL)
		return (ENOMEM);

	LIST_INSERT_HEAD(&(h[slot].head), fle, next);
	bcopy(r, &fle->r, sizeof(struct flow_rec));
	fle->bytes = plen;
	fle->packets = 1;

	getmicrouptime(&tv);
	fle->first = fle->last = tv.tv_sec;

	/* First we do route table lookup on destination address. So we can
	 * fill in out_ifx, dst_mask, nexthop, and dst_as in future releases.
	 */
	bzero((caddr_t)&ro, sizeof(ro));
	sin = (struct sockaddr_in *)&ro.ro_dst;
	sin->sin_len = sizeof(*sin);
	sin->sin_family = AF_INET;
	sin->sin_addr = fle->r.r_dst;
#ifdef	RTF_PRCLONING	/* disappeared in CURRENT */
	rtalloc_ign(&ro, RTF_CLONING|RTF_PRCLONING);
#else
	rtalloc_ign(&ro, RTF_CLONING);
#endif
	if (ro.ro_rt != NULL) {
		struct rtentry *rt = ro.ro_rt;

		fle->fle_o_ifx = rt->rt_ifp->if_index;

		if (rt->rt_flags & RTF_GATEWAY &&
		    rt->rt_gateway->sa_family == AF_INET)
			fle->next_hop =
			    ((struct sockaddr_in *)(rt->rt_gateway))->sin_addr;

		if (rt_mask(rt))
			fle->dst_mask =
			    bit_count(((struct sockaddr_in *)rt_mask(rt))->sin_addr.s_addr);
		else if (rt->rt_flags & RTF_HOST)
			/* Give up. We can't determine mask :( */
			fle->dst_mask = 32;

		RTFREE(ro.ro_rt);
	}

	/* Do route lookup on source address, to fill
	 * in src_mask
	 */

	bzero((caddr_t)&ro, sizeof(ro));
	sin = (struct sockaddr_in *)&ro.ro_dst;
	sin->sin_len = sizeof(*sin);
	sin->sin_family = AF_INET;
	sin->sin_addr = fle->r.r_src;
#ifdef	RTF_PRCLONING	/* disappeared in CURRENT */
	rtalloc_ign(&ro, RTF_CLONING|RTF_PRCLONING);
#else
	rtalloc_ign(&ro, RTF_CLONING);
#endif
	if (ro.ro_rt != NULL) {
		struct rtentry *rt = ro.ro_rt;

		if (rt_mask(rt))
			fle->src_mask =
			    bit_count(((struct sockaddr_in *)rt_mask(rt))->sin_addr.s_addr);
		else if (rt->rt_flags & RTF_HOST)
			/* Give up. We can't determine mask :( */
			fle->src_mask = 32;

		RTFREE(ro.ro_rt);
	}

	return (0);
}

static __inline int
make_flow_rec(struct mbuf **m, int *plen, struct flow_rec *r,
    u_int8_t *tcp_flags, u_int16_t i_ifx)
{
	register struct ip *ip;
	int hlen;
	int error = 0;

	ip = mtod(*m, struct ip*);

	/* check version */
#ifdef	IP_VHL_V			/* keep compatibility with 4.x */
	if (IP_VHL_V(ip->ip_vhl) != IPVERSION)
#else
	if (ip->ip_v != IPVERSION)
#endif
		return (EINVAL);

	/* verify min header length */
#ifdef	IP_VHL_HL 			/* keep compatibility with 4.x */
	hlen = IP_VHL_HL(ip->ip_vhl) << 2;
#else
	hlen = ip->ip_hl << 2;
#endif
	if (hlen < sizeof(struct ip))
		return (EINVAL);

	r->r_src = ip->ip_src;
	r->r_dst = ip->ip_dst;

	/* save packet length */
	*plen = ntohs(ip->ip_len);

	r->r_ip_p = ip->ip_p;
	r->r_tos = ip->ip_tos;

	if ((*m)->m_pkthdr.rcvif)
		r->r_i_ifx = (*m)->m_pkthdr.rcvif->if_index;
	else
		r->r_i_ifx = i_ifx;

	/*
	 * XXX NOTE: only first fragment of fragmented TCP, UDP and
	 * ICMP packet will be recorded with proper s_port and d_port.
	 * Following fragments will be recorded simply as IP packet with
	 * ip_proto = ip->ip_p and s_port, d_port set to zero.
	 * I know, it looks like bug. But I don't want to re-implement
	 * ip packet assebmling here. Anyway, (in)famous trafd works this way -
	 * and nobody complains yet :)
	 */
	if(ip->ip_off & htons(IP_OFFMASK))
		return (0);

	/* skip IP header */
	m_adj(*m, hlen);

	switch(r->r_ip_p) {
	case IPPROTO_TCP:
	{
		register struct tcphdr *tcp;

		/* verify that packet is not truncated */
		if (CHECK_MLEN(*m, sizeof(struct tcphdr)))
			ERROUT(EINVAL);

		if (CHECK_PULLUP(*m, sizeof(struct tcphdr)))
			ERROUT(ENOBUFS);

		tcp = mtod(*m, struct tcphdr*);
		r->r_sport = tcp->th_sport;
		r->r_dport = tcp->th_dport;
		*tcp_flags = tcp->th_flags;
		break;
	}
	case IPPROTO_UDP:
		/* verify that packet is not truncated */
		if (CHECK_MLEN(*m, sizeof(struct udphdr)))
			ERROUT(EINVAL);

		if (CHECK_PULLUP(*m, sizeof(struct udphdr)))
			ERROUT(ENOBUFS);

		r->r_ports = *(mtod(*m, u_int32_t *));
		break;
	}

done:
	return (error);
}

/*
 * Non-static functions called from ng_netflow.c
 *
 */

/* Create empty hash array */
int
hash_init(priv_p priv)
{
	struct flow_chunk	*chunk;

	MALLOC(priv->hash, struct flow_hash *,
	    NBUCKETS * sizeof(struct flow_hash),
	    M_NETFLOW, M_NOWAIT);
	if (priv->hash == NULL)
		return (ENOMEM);

	bzero(priv->hash, NBUCKETS * sizeof(struct flow_hash));

	MALLOC(chunk, struct flow_chunk *, sizeof(struct flow_chunk),
	    M_NETFLOW, M_NOWAIT);
	if (chunk == NULL)
		return (ENOMEM);

	bzero(chunk, sizeof(struct flow_chunk));
	LIST_INSERT_HEAD(&priv->cache, chunk, entries);

	return (0);
}

/* Free all flow cache memory. Called from ng_rmnode, and thus non-static. */
void
hash_flush(priv_p priv)
{
	register struct flow_chunk *chunk;

	/* free hash memory */
	if (priv->hash)
		FREE(priv->hash, M_NETFLOW);

	/* walk down through cache flow and free memory */
	while (!LIST_EMPTY(&priv->cache)) {
	     chunk = LIST_FIRST(&priv->cache);
	     LIST_REMOVE(chunk, entries);
	     FREE(chunk, M_NETFLOW);
	}

}

/*
 * Insert packet from &m into flow cache
 */
int
flow_add(priv_p priv, struct mbuf **m, iface_p iface)
{
	struct flow_hash	*h = priv->hash;
	struct ng_netflow_info	*info = &(priv->info);
	register struct flow_entry *fle;
	struct flow_rec		r;
	int			plen;
	int			error = 1;
	u_int32_t		slot;
	u_int8_t		tcp_flags = 0;
	struct timeval		tv;

	/* Try to fill *rec */
	bzero(&r, sizeof(r));
	if ((error = make_flow_rec(m, &plen, &r, &tcp_flags,
	    iface->info.if_index)))
		return (error);

	slot = ip_hash(&r);
	fle = hash_lookup(h, slot, &r);

	getmicrouptime(&tv);

	if (fle) {
		fle->bytes += plen;
		fle->packets ++;
		fle->tcp_flags |= tcp_flags;
		fle->last = tv.tv_sec;

		if (tcp_flags & TH_FIN || tcp_flags & TH_RST ||
		    fle->bytes >= (UINT_MAX - IF_MAXMTU)) {
			if ((error = export_add(priv, fle)) != 0)
				printf("ng_netflow: export_add() failed %d\n",
				    error);
			priv->info.records --;
			LIST_REMOVE(fle, next);
			free_flow(fle);
		}
	} else {
		/*
		 * End of slot was reached. We need to allocate
		 * new record.
		 */
		if ((error = hash_insert(priv, slot, &r, plen, iface)))
	                return (error);
	        info->records ++;
	}

	if (priv->last_expire < tv.tv_sec &&
	    priv->export != NULL)
		return expire(priv);

	return (0);
}

/*
 * Return records from cache
 * XXX: matching particular IP should be done in kernel, here.
 */

int
flow_show(priv_p priv, struct ng_mesg *resp)
{
	register struct flow_entry *fle;
	struct flows	*data;
	static struct flow_chunk *chunk = NULL;
	static int i = 0;

	data = (struct flows *)resp->data;
	data->nentries = 0;

	if (chunk == NULL)
		chunk = LIST_FIRST(&priv->cache);

	if (i == -1) {	/* previous run reached end of cache */
		i = 0;
		chunk = LIST_FIRST(&priv->cache);
		return (0);
	}

	/* We will transfer not more than NREC_AT_ONCE. More data
	 * will come in next message. Current stop point is stored
	 * in static variables.
	 */
	while (chunk != NULL) {
		fle = &(chunk->chunk[i]);
		for(; i < CHUNK_SIZE; i++, fle++) {
			if (fle->packets != 0) {
				bcopy(fle, &(data->entries[data->nentries]), sizeof(*fle));
				data->nentries ++;
				if (data->nentries == NREC_AT_ONCE)
					return (0);
			}
		}
		i = 0;
		chunk = LIST_NEXT(chunk, entries);
     	}

	/* Next flow_show() must return empty response indicating that
	 * end of cache is reached */

	i = -1;

	return (0);
}

/*
 * We have full datagram in privdata. Send it to export.
 */

static int
export_send(priv_p priv)
{
	struct netflow_v5_header *header = &priv->header;
	struct timeval tv;
	struct timespec ts;
	struct mbuf *m;
	struct m_tag *mt;
	int mlen;
	int error = 0;
#if (NG_ABI_VERSION == 2)
	void *dummy = NULL;
#endif

	getmicrouptime(&tv);
	header->sys_uptime = htonl(tv.tv_sec);

	getnanotime(&ts);
	header->unix_secs  = htonl(ts.tv_sec);
	header->unix_nsecs = htonl(ts.tv_nsec);

	/* Flow sequence contains number of first record */
	header->flow_seq = htonl(priv->flow_seq - header->count);

	mlen = sizeof(struct netflow_v5_header) +
	    sizeof(struct netflow_v5_record) * header->count;

	header->count = htons(header->count);
	if ((m = m_devget((caddr_t)header, mlen, 0, NULL, NULL)) == NULL) {
		printf("ng_netflow: m_devget() failed, losing export dgram\n");
		header->count = 0;
		return(ENOBUFS);
	}

	/* We have only one type of tags now, so allocate tag of zero length */
	if ((mt = m_tag_alloc(NGM_NETFLOW_COOKIE, 0, 0, M_NOWAIT)) == NULL)
		printf("ng_netflow: tag allocation failed\n");
	else
		m_tag_prepend(m, mt);

	header->count = 0;

#if (NG_ABI_VERSION == 2)
	NG_SEND_DATA(error, priv->export, m, dummy);
	if (error)
		NG_FREE_DATA(m, dummy);
#else
	NG_SEND_DATA_ONLY(error, priv->export, m);
	if (error)
		NG_FREE_M(m);
#endif

	return (error);
}


/*
 * Create export datagram.
 */

static int
export_add(priv_p priv, struct flow_entry *fle)
{
	struct netflow_v5_header *header = &priv->header;
	struct netflow_v5_record *rec;

	/* fle == NULL indicates that expiry run has completed and something
	 * must be sent. However, people working with Cisco routers
	 * say that Cisco does not send empty datagrams. We will pretend
	 * to be Cisco by default.
	 */
	if (fle == NULL) {
#ifndef	EMPTY_EXPORTS
		if (header->count == 0)
			return (0);
		else
#endif
			return export_send(priv);
	}

	if (header->count == 0 ) {	/* first record */
		rec = &priv->r[0];
		header->count = 1;
	} else {			/* continue filling datagram */
		rec = &priv->r[header->count];
		header->count ++;
	}

	/* XXX: this must be optimized */
	rec->src_addr = fle->r.r_src.s_addr;
	rec->dst_addr = fle->r.r_dst.s_addr;
	rec->next_hop = fle->next_hop.s_addr;
	rec->i_ifx    = htons(fle->fle_i_ifx);
	rec->o_ifx    = htons(fle->fle_o_ifx);
	rec->packets  = htonl(fle->packets);
	rec->octets   = htonl(fle->bytes);
	rec->first    = htonl(fle->first);
	rec->last     = htonl(fle->last);
	rec->s_port   = fle->r.r_sport;
	rec->d_port   = fle->r.r_dport;
	rec->flags    = fle->tcp_flags;
	rec->prot     = fle->r.r_ip_p;
	rec->tos      = fle->r.r_tos;
	rec->dst_mask = fle->dst_mask;
	rec->src_mask = fle->src_mask;

	priv->flow_seq ++;

	if (header->count == NETFLOW_V5_MAX_RECORDS) /* end of dgram */
		return export_send(priv);

	return (0);	
}

/*
 * Second implementation of flow expiry.
 * Some things surely can be optimized.
 */

int
expire(priv_p priv)
{
	struct flow_chunk		*chunk;
	register struct flow_entry	*fle;
	register int i;
	struct timeval tv;
	int error = 0;

	getmicrouptime(&tv);

	LIST_FOREACH(chunk, &(priv->cache), entries) {
		fle = chunk->chunk;
		if (chunk->flows == 0)	/* empty chunk */
			continue;
		for (i = 0; i < CHUNK_SIZE; i++, fle++) {
			if (fle->packets != 0 &&
			    ( tv.tv_sec - fle->first > priv->info.active_timeout ||
			      tv.tv_sec - fle->last > priv->info.inactive_timeout))
			{

				if ((error = export_add(priv, fle)) != 0)
					printf("ng_netflow: export_add() failed: %u\n",
					    error);
				priv->info.records --;
				LIST_REMOVE(fle, next);
				free_flow(fle);
			}
		}
     	}

	/* Expiry run has finished. Now push export datagram. */
	export_add(priv, NULL);

	/* schedule next expire */
	priv->last_expire = tv.tv_sec;

	return (0);
}
